<?php

/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */

/**
 * Server commands for our PHP implementation of the XML-RPC protocol
 *
 * This is a PEAR-ified version of Useful inc's XML-RPC for PHP.
 * It has support for HTTP transport, proxies and authentication.
 *
 * PHP versions 4 and 5
 *
 * LICENSE: License is granted to use or modify this software
 * ("XML-RPC for PHP") for commercial or non-commercial use provided the
 * copyright of the author is preserved in any distributed or derivative work.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESSED OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @category   Web Services
 * @package    XML_RPC
 * @author     Edd Dumbill <edd@usefulinc.com>
 * @author     Stig Bakken <stig@php.net>
 * @author     Martin Jansen <mj@php.net>
 * @author     Daniel Convissor <danielc@php.net>
 * @author     Edgar Soldin <edgar.soldin@web.de>
 * @copyright  1999-2001 Edd Dumbill, 2001-2005 The PHP Group,
 * @copyright  Multicall & Authentification 2005 Edgar Soldin 
 * @version    CVS: $Id: Server.php,v 1.30 2005/09/07 04:06:20 danielc Exp $
 * @link       http://pear.php.net/package/XML_RPC
 */


/**
 * Pull in the XML_RPC class
 */
require_once 'XML/RPC.php';


/**
 * signature for system.listMethods: return = array,
 * parameters = a string or nothing
 * @global array $GLOBALS['XML_RPC_Server_listMethods_sig']
 */
$GLOBALS['XML_RPC_Server_listMethods_sig'] = array(
    array($GLOBALS['XML_RPC_Array'],
          $GLOBALS['XML_RPC_String']
    ),
    array($GLOBALS['XML_RPC_Array'])
);

/**
 * docstring for system.listMethods
 * @global string $GLOBALS['XML_RPC_Server_listMethods_doc']
 */
$GLOBALS['XML_RPC_Server_listMethods_doc'] = 'This method lists all the'
        . ' methods that the XML-RPC server knows how to dispatch';

/**
 * signature for system.methodSignature: return = array,
 * parameters = string
 * @global array $GLOBALS['XML_RPC_Server_methodSignature_sig']
 */
$GLOBALS['XML_RPC_Server_methodSignature_sig'] = array(
    array($GLOBALS['XML_RPC_Array'],
          $GLOBALS['XML_RPC_String']
    )
);

/**
 * docstring for system.methodSignature
 * @global string $GLOBALS['XML_RPC_Server_methodSignature_doc']
 */
$GLOBALS['XML_RPC_Server_methodSignature_doc'] = 'Returns an array of known'
        . ' signatures (an array of arrays) for the method name passed. If'
        . ' no signatures are known, returns a none-array (test for type !='
        . ' array to detect missing signature)';

/**
 * signature for system.methodHelp: return = string,
 * parameters = string
 * @global array $GLOBALS['XML_RPC_Server_methodHelp_sig']
 */
$GLOBALS['XML_RPC_Server_methodHelp_sig'] = array(
    array($GLOBALS['XML_RPC_String'],
          $GLOBALS['XML_RPC_String']
    )
);

/**
 * docstring for methodHelp
 * @global string $GLOBALS['XML_RPC_Server_methodHelp_doc']
 */
$GLOBALS['XML_RPC_Server_methodHelp_doc'] = 'Returns help text if defined'
        . ' for the method passed, otherwise returns an empty string';

/**
 * dispatch map for the automatically declared XML-RPC methods.
 * @global array $GLOBALS['XML_RPC_Server_dmap']
 */
$GLOBALS['XML_RPC_Server_dmap'] = array (
    'system.listMethods' => array (
        'function' => 'XML_RPC_Server.system_listMethods', 
        'signature' => $GLOBALS['XML_RPC_Server_listMethods_sig'], 
        'docstring' => $GLOBALS['XML_RPC_Server_listMethods_doc']
    ), 
    'system.methodHelp' => array (
        'function' => 'XML_RPC_Server.system_methodHelp', 
        'signature' => $GLOBALS['XML_RPC_Server_methodHelp_sig'], 
        'docstring' => $GLOBALS['XML_RPC_Server_methodHelp_doc']
    ), 
    'system.methodSignature' => array (
        'function' => 'XML_RPC_Server.system_methodSignature', 
        'signature' => $GLOBALS['XML_RPC_Server_methodSignature_sig'], 
        'docstring' => $GLOBALS['XML_RPC_Server_methodSignature_doc']
    ), 
    'system.echo' => array (
        'function' => 'XML_RPC_Server.system_echo',
        'signature' => array (
             array ($XML_RPC_Array, $XML_RPC_Array), 
             array ($XML_RPC_Int, $XML_RPC_Int), 
             array ($XML_RPC_String, $XML_RPC_String), 
             array ($XML_RPC_Boolean, $XML_RPC_Boolean)
        ), 
        'docstring' => 'Simply echoes the parameter of the message (for test purposes).'
    ), 
    'system.error' => array (
        'function' => 'XML_RPC_Server.system_error', 
        'signature' => array (), 
        'docstring' => 'Returns a bogus error for test purposes.'
    ), 
    'system.multiCall' => array (
        'function' => 'XML_RPC_Server.system_multiCall', 
        'signature' => array (array ($XML_RPC_Array, $XML_RPC_Array)), 
        'docstring' => 'Executes multiple methods in sequence and returns the results '.
        '(implements http://www.xmlrpc.com/discuss/msgReader$1208).'
    ), 
    'system.time' => array (
        'function' => 'XML_RPC_Server.system_time', 
        'signature' => array (array ()), 
        'docstring' => 'Return the server time as unix time stamp')
    );

/**
 * @global string $GLOBALS['XML_RPC_Server_debuginfo']
 */
$GLOBALS['XML_RPC_Server_debuginfo'] = '';


/**
 * Lists all the methods that the XML-RPC server knows how to dispatch
 *
 * @param object a XML_RPC_Server object
 * @param object a XML_RPC_Message object
 *
 * @return object  a new XML_RPC_Response object
 */
function XML_RPC_Server_listMethods($server, $m)
{
    return $server->system_methodHelp($m);
}

/**
 * Returns an array of known signatures (an array of arrays)
 * for the given method
 *
 * If no signatures are known, returns a none-array
 * (test for type != array to detect missing signature)
 *
 * @param object a XML_RPC_Server object
 * @param object a XML_RPC_Message object
 *
 * @return object  a new XML_RPC_Response object
 */
function XML_RPC_Server_methodSignature($server, $m)
{
    return $server->system_methodSignature($m);
}

/**
 * Returns help text if defined for the method passed, otherwise returns
 * an empty string
 *
 * @param object a XML_RPC_Server object
 * @param object a XML_RPC_Message object
 * 
 * @return object  a new XML_RPC_Response object
 */
function XML_RPC_Server_methodHelp($server, $m)
{
    return $server->system_methodHelp($m);
}

/**
 * @return void
 */
function XML_RPC_Server_debugmsg($m)
{
    global $XML_RPC_Server_debuginfo;
    $XML_RPC_Server_debuginfo = $XML_RPC_Server_debuginfo . $m . "\n";
}


/**
 * A server for receiving and replying to XML RPC requests
 *
 * <code>
 * $server = new XML_RPC_Server(
 *     array(
 *         'isan8' =>
 *             array(
 *                 'function' => 'is_8',
 *                 'signature' =>
 *                      array(
 *                          array('boolean', 'int'),
 *                          array('boolean', 'int', 'boolean'),
 *                          array('boolean', 'string'),
 *                          array('boolean', 'string', 'boolean'),
 *                      ),
 *                 'docstring' => 'Is the value an 8?'
 *             ),
 *     ),
 *     1,
 *     0
 * ); 
 * </code>
 *
 * @category   Web Services
 * @package    XML_RPC
 * @author     Edd Dumbill <edd@usefulinc.com>
 * @author     Stig Bakken <stig@php.net>
 * @author     Martin Jansen <mj@php.net>
 * @author     Daniel Convissor <danielc@php.net>
 * @copyright  1999-2001 Edd Dumbill, 2001-2005 The PHP Group
 * @version    Release: @package_version@
 * @link       http://pear.php.net/package/XML_RPC
 * 
 * @todo       make ALL system_methods msg->send_encoding aware
 */
class XML_RPC_Server
{
    /**
     * The dispatch map, listing the methods this server provides.
     * @var array
     */
    var $dmap = array();

    /**
     * The present response's encoding
     * @var string
     * @see XML_RPC_Message::getEncoding()
     */
    var $encoding = '';

    /**
     * Debug mode (0 = off, 1 = on)
     * @var integer
     */
    var $debug = 0;

    /**
     * The response's HTTP headers
     * @var string
     */
    var $server_headers = '';

    /**
     * The response's XML payload
     * @var string
     */
    var $server_payload = '';


    /**
     * Constructor for the XML_RPC_Server class
     *
     * @param array $dispMap   the dispatch map. An associative array
     *                          explaining each function. The keys of the main
     *                          array are the procedure names used by the
     *                          clients. The value is another associative array
     *                          that contains up to three elements:
     *                            + The 'function' element's value is the name
     *                              of the function or method that gets called.
     *                              To define a class' method: 'class::method'.
     *                            + The 'signature' element (optional) is an
     *                              array describing the return values and
     *                              parameters
     *                            + The 'docstring' element (optional) is a
     *                              string describing what the method does
     * @param int $serviceNow  should the HTTP response be sent now?
     *                          (1 = yes, 0 = no)
     * @param int $debug       should debug output be displayed?
     *                          (1 = yes, 0 = no)
     *
     * @return void
     */
    function XML_RPC_Server($dmap = array (), $serviceNow = 1, $debug = 0)
    {
    
        // set debug mode
        $this->debug($debug);
    
        // userdefined dmap entries overwrite predefined ones
        $this->dmap = array_merge($GLOBALS['XML_RPC_Server_dmap'], $dmap);
    
        // payload is created while servicing
        if ($serviceNow) {
          $this->service();
        }
    
    }

    /**
     * En-/Disable debug mode
     * 
     * @return boolean 
     */
    function debug($state = null)
    {
        // keeping $this->debug for backward compatibility
        if (!is_null($state)) {
            $this->config['debug'] = $this->debug = $state;
        }
        return (boolean)$this->config['debug'];
    }

    /**
     * Get/Set config vars.
     * Currently these are: 
     * 
     * @return boolean 
     */
    function config($config = array ())
    {
        // add to config array and return it
        if (is_array($config)) {
            return $this->config = array_merge((array) $this->config, $config);
        }
        // get config for a given key or return null if not set
        elseif (is_string($key = $config)) {
            return $this->config[$key];
        }
        // return config by default
        else {
            return $this->config;
        }
    }

    /**
     * @return string  the debug information if debug debug mode is on
     */
    function serializeDebug()
    {
        global $XML_RPC_Server_debuginfo, $HTTP_RAW_POST_DATA;

        if ($this->debug) {
            XML_RPC_Server_debugmsg('vvv POST DATA RECEIVED BY SERVER vvv' . "\n"
                                    . $HTTP_RAW_POST_DATA
                                    . "\n" . '^^^ END POST DATA ^^^');
        }

        if ($XML_RPC_Server_debuginfo != '') {
            return "<!-- PEAR XML_RPC SERVER DEBUG INFO:\n\n"
                   . preg_replace('/-(?=-)/', '- ', $XML_RPC_Server_debuginfo)
                   . "-->\n";
        } else {
            return '';
        }
    }


    /**
     * Authenticates if 'auth' params are defined
     *
     * @return boolean
     */
    function authenticate() {
        $config = $this->config('auth');
        // empty config returns always true, else check auth   
        return empty ($config) || (array_key_exists($_SERVER['PHP_AUTH_USER'], $config) 
                                    && $_SERVER['PHP_AUTH_PW'] == $config[$_SERVER['PHP_AUTH_USER']]);
    }

    /**
     * Sends the response
     *
     * The encoding and content-type are determined by
     * XML_RPC_Message::getEncoding()
     *
     * @return void
     *
     * @uses XML_RPC_Server::createServerPayload(),
     *       XML_RPC_Server::createServerHeaders()
     */
    function service() 
    {
        if ($this->authenticate()) {
    
            if (!$this->server_payload) {
                $this->createServerPayload();
            }
            if (!$this->server_headers) {
                $this->createServerHeaders();
            }
            header($this->server_headers);

        
            // @TODO: should this be done when serializing values?
            if ($this->encoding == 'UTF-8') {
                $this->server_payload = utf8_encode($this->server_payload);
            }
            
            print $this->server_payload;
        
        } 
        else 
        {
            header('WWW-Authenticate: Basic realm="HTTP Authentication"');
            header('HTTP/1.0 401 Unauthorized');
            echo "Access not allowed.\n";
        }
    }

    /**
     * Generates the payload and puts it in the $server_payload property
     *
     * @return void
     *
     * @uses XML_RPC_Server::parseRequest(), XML_RPC_Server::$encoding,
     *       XML_RPC_Response::serialize(), XML_RPC_Server::serializeDebug()
     */
    function createServerPayload()
    {
        $r = $this->parseRequest();
        $this->server_payload = '<?xml version="1.0" encoding="'
                              . $this->encoding . '"?>' . "\n"
                              . $this->serializeDebug()
                              . $r->serialize();
    }

    /**
     * Determines the HTTP headers and puts them in the $server_headers
     * property
     *
     * @return boolean  TRUE if okay, FALSE if $server_payload isn't set.
     *
     * @uses XML_RPC_Server::createServerPayload(),
     *       XML_RPC_Server::$server_headers
     */
    function createServerHeaders()
    {
        if (!$this->server_payload) {
            return false;
        }
        $this->server_headers = 'Content-Length: '
                              . strlen($this->server_payload) . "\r\n"
                              . 'Content-Type: text/xml;'
                              . ' charset=' . $this->encoding;
        return true;
    }

    /**
     * verifies the given messages signature against the defined ones in dmap
     * 
     * @return PEAR_Error on error
     */
    function verifySignature($msg) {
        global $XML_RPC_err, $XML_RPC_str;
        $method = $msg->method();
        $sigs_def = $this->dmap[$method]['signature'];
    
        // empty sig, .. nothing to check
        if (!isset ($sigs_def) || !count($sigs_def)) 
        {
            //echo("$method/EMPTY\n");
            return;
        }        
    
        for ($i = 0; $i < $sigs_def_count = count($sigs_def); $i ++) {
            $sig_ok = false;
            $sig_def = $sigs_def[$i];
            
            // count of params differs, proceed w/ next sig
            if (count($sig_def) - 1 != $msg->getNumParams()) {
                //echo("$method/WRONG COUNT\n");
                continue;
            }
            
            // empty paramlist in sigs_def meets no params -> valid
            if (count($sig_def) - 1 < 1 && $msg->getNumParams() < 1) {
                return;
            }
        
            for ($j = 1; $j < $max = count($sig_def); $j ++) 
            {
                $sig_ptype = $sig_def[$j];
                $msg_p = $msg->getParam($j -1);
        
                // detect msg parameters type
                if ($msg_p->kindOf() == 'scalar') {
                    $msg_ptype = $msg_p->scalartyp();
                } 
                else {
                    $msg_ptype = $msg_p->kindOf();
                }
                // compare types and react
                if ($sig_ptype == $msg_ptype) {
                    // this param is ok check next
                    // or if last exit method w/o error
                    $sig_ok = true;
                    if ($j == count($sig_def) - 1) {
                        //echo("$method/MATCH\n");
                        return;
                    }
                } 
                else {
                    // this param is wrong, check next sig
                    $sig_ok = false;
                    break;
                }
            }
        }
    
        //echo("$method/FAIL\n");
    
        // not sig_ok, return a nice error message
        // a nice description string for sigs_def
        unset ($sigs_desc);
        foreach ($sigs_def as $key => $sig_def) 
        {
            $sigs_desc .= ($sigs_desc ? ', ' : '')."$key {";
            unset ($sig_desc);
            // dont add the first (return value) to list
            for ($k = 1; $k < count($sig_def); $k ++) {
                $sig_desc .= ($sig_desc ? ', ' : '').$sig_def[$k];
            }
            $sigs_desc .= "$sig_desc}";
        }
    
        // description of given msg param types
        for ($k = 0; $k < $max = $msg->getNumParams(); $k ++) 
        {
            $p = $msg->getParam($k);
            if (($type = $p->kindOf()) == 'scalar') {
                $type = $p->scalartyp();
            }
            $msg_desc .= ($msg_desc ? ', ' : '').$type;
        }
        $msg_desc = "{". ($msg_desc ? $msg_desc : 'none')."}";
        return PEAR :: raiseError(basename(__FILE__).'::verifySignature() for ' .
                'method \''.$msg->method()."' failed because the parameters you " .
                "gave $msg_desc do not match one of "."defined parameter " .
                "signatures ($sigs_desc).");
    
    }


    /**
     * @return object  a new XML_RPC_Response object
     *
     * @uses XML_RPC_Message::getEncoding(), XML_RPC_Server::$encoding
     */
    function parseRequest($data = '')
    {
        global $XML_RPC_xh, $HTTP_RAW_POST_DATA,
                $XML_RPC_err, $XML_RPC_str, $XML_RPC_errxml,
                $XML_RPC_defencoding, $XML_RPC_Server_dmap;

        if ($data == '') {
            $data = $HTTP_RAW_POST_DATA;
        }

        $this->encoding = XML_RPC_Message::getEncoding($data);
        $parser_resource = xml_parser_create($this->encoding);
        $parser = (int) $parser_resource;

        $XML_RPC_xh[$parser] = array();
        $XML_RPC_xh[$parser]['cm']     = 0;
        $XML_RPC_xh[$parser]['isf']    = 0;
        $XML_RPC_xh[$parser]['params'] = array();
        $XML_RPC_xh[$parser]['method'] = '';
        $XML_RPC_xh[$parser]['stack'] = array();    
        $XML_RPC_xh[$parser]['valuestack'] = array();   

        $plist = '';

        // decompose incoming XML into request structure

        xml_parser_set_option($parser_resource, XML_OPTION_CASE_FOLDING, true);
        xml_set_element_handler($parser_resource, 'XML_RPC_se', 'XML_RPC_ee');
        xml_set_character_data_handler($parser_resource, 'XML_RPC_cd');
        if (!xml_parse($parser_resource, $data, 1)) {
            // return XML error as a faultCode
            $r = new XML_RPC_Response(0,
                                      $XML_RPC_errxml+xml_get_error_code($parser_resource),
                                      sprintf('XML error: %s at line %d',
                                              xml_error_string(xml_get_error_code($parser_resource)),
                                              xml_get_current_line_number($parser_resource)));
            xml_parser_free($parser_resource);
        } elseif ($XML_RPC_xh[$parser]['isf']>1) {
            $r = new XML_RPC_Response(0,
                                      $XML_RPC_err['invalid_request'],
                                      $XML_RPC_str['invalid_request']
                                      . ': '
                                      . $XML_RPC_xh[$parser]['isf_reason']);
            xml_parser_free($parser_resource);
        } else {
            xml_parser_free($parser_resource);
            $m = new XML_RPC_Message($XML_RPC_xh[$parser]['method']);
            $m->setSendEncoding($this->encoding);
            // now add parameters in
            for ($i = 0; $i < sizeof($XML_RPC_xh[$parser]['params']); $i++) {
                //print '<!-- ' . $XML_RPC_xh[$parser]['params'][$i]. "-->\n";
                $plist .= "$i - " . var_export($XML_RPC_xh[$parser]['params'][$i], true) . " \n";
                $m->addParam($XML_RPC_xh[$parser]['params'][$i]);
            }

            if ($this->debug) {
                XML_RPC_Server_debugmsg($plist);
            }

            // now to deal with the method
            $methName = $XML_RPC_xh[$parser]['method'];
            $dmap = $this->dmap;
            $function = $dmap[$methName]['function'];
      
            if (!$function) {
                $r = new XML_RPC_Response(0, $XML_RPC_err['unknown_method'], 
                    "Serverside dmap method key '$methName' results in an empty " .
                    "function. Server dmap configuration problem? Check upper " .
                    "and lower case as the method's name is case sensitive.");
            }
            else {
                // dispatch procedure call
                $r = $this->execute($dmap[$methName]['function'], $m);
            }
        }
        
        return $r;
    }

    /**
     * Executes the function defined in dmap and return result in response form 
     * or return 'method not found' response error. 
     * Note: 
     *   It is expected that the called function/method returns a valid response
     *   object.
     * 
     * Supported are:
     * <ul>
     *  <li>
     *   Static class functions in form of 'Class::function'
     *  </i>
     *  <li>
     *   Server objects methods in form of 'Classname.Methodname', whilst classname 
     *   must be the classname of the XML_RPC_Server object serverside
     *  </li>
     *  <li>
     *   Static functions in global namespace in form of 'function'
     *  </li>
     * </ul>
     *
     * @return object  a new XML_RPC_Response object
     */
    function execute($function, $msg) {
        global $XML_RPC_err, $XML_RPC_str;
        
        // if signature does not match
        if (PEAR :: isError($err = $this->verifySignature($msg, $sig)))
        {
            return new XML_RPC_Response(0, $XML_RPC_err['incorrect_params'], $XML_RPC_str['incorrect_params'].': '.$err->getMessage());
        }
        
        if (!preg_match("/:|\./", $function) && is_callable($function, false)) 
        {
            //echo "<p>Function '$function'</p>";
            $r = call_user_func($function, $msg);
        }
        elseif (preg_match("/^([^\.]+)\.([^\.]+)$/i", $function, $matches) 
                && is_callable(array ($object = $matches[1], $method = $matches[2]), false) 
                && preg_match("/^".preg_quote($object)."$/i", get_class($this))) 
        {
    
            //echo "<p>ServerObject '$function' '".get_class($this)."'</p>";
            $r = call_user_func(array ($this, $method), $msg);
        }
        elseif (preg_match("/^([\w]+)(?:::)([\w]+)$/i", $function, $matches) 
                && class_exists($class = $matches[1]) 
                && in_array(strtolower($classfunction = $matches[2]), get_class_methods($class)) ) 
        {
    
            //echo "<p>StaticObject '$function'</p>";
            $r = call_user_func(array($class,$classfunction),$msg);
        } 
        else 
        {
    
            //echo "<p>Unknown '$function'</p>";      
            $r = new XML_RPC_Response(0, $XML_RPC_err['unknown_method'], 'Serverside method '. ($this->debug() ? "'$function' " : "'enable serverside debug for the name' ").'unknown.'.' Server dmap configuration problem?');
        }
    
        if (!is_a($r, 'XML_RPC_Response')) 
        {
            return new XML_RPC_Response(0, $XML_RPC_err['not_response_object'], $XML_RPC_str['not_response_object']);
        } 
        else 
        {
            return $r;
        }
    }
  
    /**
     * Lists all the methods that the XML-RPC server knows how to dispatch
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_listMethods($server) {
        global $XML_RPC_err, $XML_RPC_str;
    
        $v = new XML_RPC_Value();
        $dmap = is_a($server,'XML_RPC_Server') ? $server->dmap : $this->dmap;
        $outAr = array ();
        for (reset($dmap); list ($key, $val) = each($dmap);) {
            $outAr[] = new XML_RPC_Value($key, 'string');
        }
    
        $v->addArray($outAr);
        return new XML_RPC_Response($v);
    }

    /**
     * Returns an array of known signatures (an array of arrays)
     * for the given method
     *
     * If no signatures are known, returns a none-array
     * (test for type != array to detect missing signature)
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_methodSignature($m) {
        global $XML_RPC_err, $XML_RPC_str, $XML_RPC_Server_dmap;
    
        $methName = $m->getParam(0);
        $methName = $methName->scalarval();
    
        $dmap = $this->dmap;
    
        //  print "<!-- ${methName} -->\n";
        if (isset ($dmap[$methName])) {
            if ($dmap[$methName]['signature']) {
                $sigs = array ();
                $thesigs = $dmap[$methName]['signature'];
                for ($i = 0; $i < sizeof($thesigs); $i ++) {
                    $cursig = array ();
                    $inSig = $thesigs[$i];
                    for ($j = 0; $j < sizeof($inSig); $j ++) {
                        $cursig[] = new XML_RPC_Value($inSig[$j], 'string');
                    }
                    $sigs[] = new XML_RPC_Value($cursig, 'array');
                }
                $r = new XML_RPC_Response(new XML_RPC_Value($sigs, 'array'));
            } 
            else {
                $r = new XML_RPC_Response(new XML_RPC_Value('undef', 'string'));
            }
        } 
        else {
            $r = new XML_RPC_Response(0, $XML_RPC_err['introspect_unknown'], $XML_RPC_str['introspect_unknown']);
        }
        return $r;
    }

    /**
     * Returns help text if defined for the method passed, otherwise returns
     * an empty string
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_methodHelp($m) {
        global $XML_RPC_err, $XML_RPC_str, $XML_RPC_Server_dmap;
    
        $methName = $m->getParam(0);
        $methName = $methName->scalarval();
    
        $dmap = $this->dmap;
    
        //  print "<!-- ${methName} -->\n";
        if (isset ($dmap[$methName])) {
            if ($dmap[$methName]['docstring']) {
                $r = new XML_RPC_Response(new XML_RPC_Value($dmap[$methName]['docstring']), 'string');
            } else {
                $r = new XML_RPC_Response(new XML_RPC_Value('', 'string'));
            }
        } 
        else {
          $r = new XML_RPC_Response(0, $XML_RPC_err['introspect_unknown'], $XML_RPC_str['introspect_unknown']);
        }
        return $r;
    }


    /**
     * Returns the given parameter (useful for testing)
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_echo($msg) {
        
        for ($i = 0; $i < $msg->getNumParams(); $i ++) {
            $foo[$i] = XML_RPC_decode($param=$msg->getParam($i), $msg->getSendEncoding());
        }
    
        // multiple params are returned as array (while current sig only allows 1 param)
        if (count($foo)>1){
            $val = new XML_RPC_Value($foo, 'array');
        }
        // one param is returned in its natural type
        elseif (count($foo)==1){
            $val = XML_RPC_encode($foo[0]=='ping'?'pong':$foo[0]); //ping is converted to pong (Easteregg)
        }
        // no param results returns a param error
        else {
            return new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], "system_echo() must be called with _one_ parameter (boolean, int, string, array)");
        }
        
        return new XML_RPC_Response($val);
    }

    /**
     * Executes multiple methods in sequence and returns the results 
     * (implements http://www.xmlrpc.com/discuss/msgReader$1208)
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_multiCall($msg) {
        $dmap = $GLOBALS['server']->dmap;
    
        $array = $msg->getParam(0);
        for ($i = 0; $i < $array->arraysize(); $i ++) {
            $details = $array->arraymem($i);
    
            if ($details->kindOf() != 'struct') 
            {
                $resp = new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], 
                  "system_multiCall() expects _only_ struct datatypes wrapped in one array.");  
            }
            elseif ($details->arraysize() >= 1) {

                // check if method name a string pointing to valid function
                if ( !is_a($method_obj = $details->structmem('methodName'), 'XML_RPC_value') 
                      || ($method_obj->kindOf() != 'scalar')
                      || !($method = $method_obj->scalarVal()) 
                      || !strlen($function = $dmap[$method]['function']) ) 
                {
                  
                  $resp = new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], 
                        "system_multiCall() method call '$method' type '".gettype($method).
                        "' resolves to an invalid function. Parameter or dmap configuration problem?");         
                }
                // check params object valid
                elseif ( ($params = $details->structmem('params')) && (!is_a($params, 'XML_RPC_Value') 
                          || $params->kindOf() != 'array') ) 
                {
                  
                  $resp = new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], 
                        "system_multiCall() method call '$function' parameters container " .
                        "is not a XML_RPC_Value type '$GLOBALS[XML_RPC_Array]'.");
                }
                // don't call multiCall recursive
                elseif ( preg_match("/\bsystem_multiCall$/i", $function) ) 
                {
                  
                  $resp = new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], 
                                "system_multiCall() must not be called recursively.");
                }
                // process params, build a message, execute procedure call
                else 
                {
                    // build array containing xml_rpc_value of each param
                    $params_list = array ();
                    for ($j = 0; isset ($params) && $j < $params->arraysize(); $j ++) {
                        $params_list[] = $param = $params->arraymem($j);
                    }
                  
                    $msg = new XML_RPC_Message($method, $params_list);
                    $msg->setSendEncoding($this->encoding);
                    $resp = $this->execute($function, $msg);
                }
    
            } 
            else 
            {
                $resp = new XML_RPC_Response(0, $GLOBALS['XML_RPC_err']['incorrect_params'], 
                                                "system_multiCall() must be called with at" . 
                                                "least the functions name as parameter 1");
            }
    
            // convert error to struct (multiCall Spec)
            if (!$resp->faultCode()) {
                $values[] = $resp->value();
            } 
            else {
                $value = new XML_RPC_Value;
                $value->addStruct( array( 'faultCode' => new XML_RPC_Value($resp->faultCode(), 
                                                                           $GLOBALS['XML_RPC_Int']), 
                                          'faultString' => new XML_RPC_Value($resp->faultString(), 
                                                                             $GLOBALS['XML_RPC_String'])));
                $values[] = $value;
            }
        }
    
        $v = new XML_RPC_Value();
        $v->addArray($values);
    
        return new XML_RPC_Response($v);
    }

    /**
     * Simply returns system time
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_time() 
    {
        $val = new XML_RPC_Value(time(), 'int');
        $resp = new XML_RPC_Response($val);
        return $resp;
    }

    /**
     * Simply returns an error for test purposes
     *
     * @return object  a new XML_RPC_Response object
     */
    function system_error() 
    {
        return new XML_RPC_Response(0, 123456789, "A error for test purposes.");
    }

    /**
     * Echos back the input packet as a string value
     *
     * @return void
     *
     * Useful for debugging.
     */
    function echoInput()
    {
        global $HTTP_RAW_POST_DATA;

        $r = new XML_RPC_Response(0);
        $r->xv = new XML_RPC_Value("'Aha said I: '" . $HTTP_RAW_POST_DATA, 'string');
        print $r->serialize();
    }
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * c-hanging-comment-ender-p: nil
 * End:
 */

?>